//
//  CustomField.swift
//  Lockdown
//
//  Created by Don McBrien on 07/04/2021.
//

import SwiftUI

struct CustomField<T: CustomFieldProtocol>: View {
   // Receives a binding ('value') to a custom type (T) and converts it to 'string'.
   // This is presented for editing in a standard SwiftUI TextField. Any kind of
   // string editing is permitted and is committed either by pressing <return> or
   // switching focus away from the TextField.
      // In the latter case, any editing is lost,'string' restored from 'store' and
      // 'value' remains unchanged.
      // In the former case, 'string' is converted to type T (if possible) and 'value'
      // is updated.
      // If conversion to T is not possible, the effect is the same as losing focus.
   // The value loaded must, by definition, be valid.
   // Additionally, it is possible to provide an optional validation closure which
   // causes editing to revert to the 'store' if the condition encoded in it is not met.
   // Type T must conform to 'CustomFieldProtocol' which requires that it can be cast
   // to a String and has a failable initialised from a String.

   @EnvironmentObject var editorEnvironment: PreferencesEditorEnvironment
   @Binding var value: T { didSet { self.string = value.simpleString } }
   var fieldValidator: ((T) -> Bool)? = nil
   
   @State private var string: String = ""
   @State private var editingInProgress = false
   @State private var store = ""
   @State private var foregroundColor: Color? = .black
   
   var body: some View {
      TextField("", text: $string,
                onEditingChanged: { isEditing in editingChanged(isEditing) },
                onCommit: { validateAndCommit() })
         .onAppear { self.string = value.simpleString }  // Load 'value' into 'string'
         .customFieldStyle(foregroundColor: foregroundColor, width: editorEnvironment.cellWidth)
   }
   
   private func editingChanged(_ isEditing: Bool) {
      self.editingInProgress = isEditing
      if editingInProgress {
         // when focus is acquired, give visual clue and disable Button in caller.
         foregroundColor = .red
         editorEnvironment.commit = false
         store = string
      } else {
         // usually after focus is lost via <enter> or switching to another TextField,
         // but sadly, not when lost by using a Button, hence the need for Editor to
         // disable the button in the calling struct (such as CustomRow).
         foregroundColor = .black
         string = store
      }
   }
   
   private func validateAndCommit() {
      // Cast to type T if possible and apply optional validator.
      foregroundColor = .black
      if let optionalValidator = fieldValidator {
         if let v = T(string),
            optionalValidator(v) {
            value = v
            store = string
         } else { string = store }
      } else {
         if let v = T(string) {
            value = v
            store = string
         } else { string = store }
      }
      editorEnvironment.commit = true
   }
}

//MARK: - Previews
struct CustomField_Previews: PreviewProvider {
   @EnvironmentObject var editor: PreferencesEditorEnvironment
   static var previews: some View {
      VStack {
         CustomField(value: .constant(123.45))
         CustomField(value: .constant(67))
      }
   }
}
